/* An omega network example using simjava
 */
package omega;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import eduni.simanim.*;
import eduni.simjava.*;

//
//  The simulation entities
//

// Src generates messages
class Src extends Sim_entity {
  private Sim_port out;
  private double delay;
  private int max;
  private int srcnum;
  private int nsrcs;
  private int workload;

  public Src(String name, int x, int y, int srcnum, int nsrcs, int max, 
	     double delay, int workload) {
    super(name,"src",x,y);
    out = new Sim_port("out", "port", Anim_port.RIGHT, 2); add_port(out);
    this.max   = max;
    this.delay = delay;
    this.srcnum = srcnum;
    this.nsrcs = nsrcs;
    this.workload = workload;
  }

  void send(int dest, double hold_time) {
    sim_schedule(out, 0.0, 1, new Integer(dest));
    if (Omega.show_msg) sim_trace(1, "S out "+dest);   
    sim_hold(hold_time);		 
  }

  public void body() {
    int i;
    
    if (workload == 0) { // 1 to all
      if (srcnum==1) {
	for(i=0; i<max; i++) {
	  send(i % nsrcs,delay);
	}    
      }
    } else if (workload==1) { // all to 1
      for(i=0; i<max; i++) {
	send(1,delay);
      }    
    } else if (workload==2) { // all to all
      for(i=0; i<max; i++) {
	send((i+srcnum) % nsrcs,delay);
      }    
    } else { // random 
      Random rdest = new Random();
      double spacing = 
	(workload==3) ? 10.0 : 
	(workload==4) ? 1.0 : 0.0;
      Sim_normal_obj rdelay = new Sim_normal_obj("Delay Object", 
						 spacing, spacing/2, srcnum);
      for(i=0; i<max; i++) {
	send(rdest.nextInt() % nsrcs, rdelay.sample());
      }    
    } 
  }
}
 

// A bucket, simply keeps a count of how many messages it receives
class Bucket extends Sim_entity {
  int count = 0;
  Sim_port in;

  public Bucket(String name, int x, int y) {
    super(name, "bucket", x, y);
    count = 0;
    in = new Sim_port("in", "port", Anim_port.LEFT, 2); add_port(in);
    add_param(new Anim_param("Count", Anim_param.VALUE, "0", 20, 10));
  }
  public void body() {
    Sim_event ev=null;
    while(true) {
      sim_get_next(ev);
      sim_hold(0.5);
      count++;
      sim_trace(1, "P "+count);
    }
  }
}


// Network switch 
class Sw extends Sim_entity {
  private Sim_port in1, in2, out1, out2;
  private double hold;
  private int state;
  private int swsize;
  private int destp;

  private static final int STRAIGHT = 0;
  private static final int CROSS    = 1;


  public Sw( String name, int x, int y, int swsize, double hold) {
    super(name,"sw",x,y);
    this.hold = hold;
    this.swsize = swsize;
    in1 = new Sim_port( "in1","port", Anim_port.LEFT, 2); add_port(in1);
    in2 = new Sim_port( "in2","port", Anim_port.LEFT, 8); add_port(in2);
    out1 = new Sim_port( "out1","port", Anim_port.RIGHT, 2); add_port(out1);
    out2 = new Sim_port( "out2","port", Anim_port.RIGHT, 8); add_port(out2);
  
    state = STRAIGHT;
    add_param(new Anim_param("State", Anim_param.STATE, "straight"));
    add_param(new Anim_param("Destp", Anim_param.VALUE, "0", -10, -10));
  }

  public void body() {
    dump_state();

    int destport = 1;

    while (true) {
      Sim_event ev = new Sim_event();

      // Get event:-
      sim_get_next(ev);

      destp = ((Integer)ev.get_data()).intValue();


      // System.out.println(get_name() + " Got " + destp);
      // Route message:- message destination is
      // 010
      // dud
      destp <<= 1;
      int mask = swsize; // 1 bit per stage
      if ((destp & mask)==0) { destport = 1; } else {destport = 2;}

      // Determine switch state
      int srcport = 1;
      if (ev.scheduled_by() == in1.get_dest()) 
	{ srcport = 1; } else { srcport = 2; }
      if (srcport == destport) { state = STRAIGHT;} else {state=CROSS;}
 
      dump_state();
      sim_hold(hold); 
      if (destport==1) {
	sim_schedule(out1, 0.0, 1, new Integer(destp));
      } else {
	sim_schedule(out2, 0.0, 1, new Integer(destp));
      }
      if (Omega.show_msg) 
	sim_trace(1, "S out"+destport+" "+destp);     // Send on result
    }
  }

  public void dump_state() {
    if(state == STRAIGHT) sim_trace(1, "P straight "+destp);
    if(state == CROSS) sim_trace(1, "P cross "+destp);
  }
}
 

// The main applet
public class Omega extends Anim_applet implements ItemListener {
  Panel inputs;
  Label msgLabel, sizeLabel, workLabel, showMsgLabel;
  TextField msgField;
  Choice sizeChoice;
  Choice workChoice;
  Checkbox showMsgBox;
  public static boolean show_msg = true;

  private int ilog2( int i )
  {
    int l = 0;
    for (l=0; (i>>=1)!=0; l++) ;
    return l;  
  }

  // initialise the gui inputs
  public void anim_init() {
    int i;
    inputs = new Panel();
    inputs.setLayout(new GridLayout(0,2));

    msgLabel = new Label("Messages:", Label.RIGHT);
    inputs.add(msgLabel);
    msgField = new TextField("5",3);
    inputs.add(msgField);

    sizeLabel = new Label("Switch size:", Label.RIGHT);
    inputs.add(sizeLabel);
    sizeChoice = new Choice();
    for(i=2; i<=32; i*=2) sizeChoice.addItem(" "+i+" ");
    sizeChoice.select(2);
    sizeChoice.addItemListener(this);;
    inputs.add(sizeChoice);
    
    workLabel = new Label("Workload:", Label.RIGHT);
    inputs.add(workLabel);
    workChoice = new Choice();
    workChoice.addItem(" One to all (broadcast) ");
    workChoice.addItem(" All to one (gather) ");
    workChoice.addItem(" All to all ");
    workChoice.addItem(" Random (10%) ");
    workChoice.addItem(" Random (50%) ");
    workChoice.addItem(" Random (100%) ");
    workChoice.select(2);
    inputs.add(workChoice);

    showMsgLabel = new Label("Show messages:", Label.RIGHT);
    inputs.add(showMsgLabel);
    showMsgBox = new Checkbox("");
    showMsgBox.setState(true);
    inputs.add(showMsgBox);


    this.add("North", inputs);
    System.out.println("Anim init done ");
  }

  // Setup the animation
  public void anim_layout() {
    int num_msg, i, swsize;
    double src_hold, sink_hold;

    System.out.println("Laying out simulation");
    num_msg = Integer.parseInt(msgField.getText());
    int swlogsize = (sizeChoice.getSelectedIndex())+1; // 1,2,3
    int workload  = (workChoice.getSelectedIndex()); 
    swsize = 1<<swlogsize; // 2,4,8
    double sw_hold = 0.05;
    src_hold = 1.0;
    sink_hold = 0.5; 
    show_msg = showMsgBox.getState();

    // Add entities
    int xsz = swlogsize;
    int ysz = swsize/2;
    for (int x = 0; x<xsz; x++) {
      for (int y=0; y<ysz; y++) {
	Sim_system.add(new Sw("Sw"+x+"."+y, 60 + x*50, 40 + y*50, 
			      swsize, sw_hold));
      }
    }

    for (int y=0; y<ysz*2; y++) {
      int swi = y / 2;
      Sim_system.add(new Bucket("Sink"+y, 60 + xsz*50, 30 + y*25));
      Sim_system.add(new Src("Src"+y,   10, 30 + y*25, 
			     y, swsize, num_msg, src_hold, workload));
    }



    // Link entities
    // Sources to 1st switches, last switches to sinks
    for (int y=0; y<ysz; y++) {
      Sim_system.link_ports("Src"+(y*2), "out", "Sw0."+y, "in1");
      Sim_system.link_ports("Src"+(y*2 +1), "out", "Sw0."+y, "in2");
      Sim_system.link_ports("Sw"+(xsz-1)+"."+y, "out1", "Sink"+(y*2), "in");
      Sim_system.link_ports("Sw"+(xsz-1)+"."+y, "out2", "Sink"+(y*2+1), "in");
    }
  
    // Internal connections
    for (int x = 0; x<xsz-1; x++) {
      for (int y=0; y<ysz; y++) {
	if (y<ysz/2) {
	  Sim_system.link_ports("Sw"+x+"."+y, "out1", 
				"Sw"+(x+1)+"."+y*2, "in1");
	  Sim_system.link_ports("Sw"+x+"."+y, "out2", 
				"Sw"+(x+1)+"."+(y*2+1), "in1");
	} else {
	  Sim_system.link_ports("Sw"+x+"."+y, "out1", 
				"Sw"+(x+1)+"."+(y-ysz/2)*2, "in2");
	  Sim_system.link_ports("Sw"+x+"."+y, "out2", 
				"Sw"+(x+1)+"."+((y-ysz/2)*2+1), "in2");
	}
      }
    }

  }

  public void itemStateChanged(ItemEvent e) {
    System.out.println("Item state changed");
    anim_relayout();
  }

  /** Allow omega to run in standalone mode */
  public static void main(String args[]) {
    Frame f = new Frame("Omege Network");
    Omega o = new Omega();
    f.add("Center", o);
    o.init();
    f.pack();
    f.show();
  }
 
}

